@@ -1,4 +1,4 @@
-// A C++ wrapper for CUDA (not quite yet)
+// A C++ wrapper for CUDA
 
 
 
@@ -13,27 +13,75 @@
 #include <stdexcept>
 #include <boost/shared_ptr.hpp>
 #include <boost/foreach.hpp>
-#include <boost/weak_ptr.hpp>
 #include <utility>
 #include <stack>
 #include <iostream>
 #include <vector>
 #include <boost/python.hpp>
+#include <boost/thread/thread.hpp>
+#include <boost/thread/tss.hpp>
+#include <boost/version.hpp>
+
+#if (BOOST_VERSION/100) < 1035
+#warning *****************************************************************
+#warning **** Your version of Boost C++ is likely too old for PyCUDA. ****
+#warning *****************************************************************
+#endif
+
+#if (CUDA_VERSION == 3000)
+#warning *****************************************************************
+#warning **** CUDA 3.0 detected. 
+#warning **** (Don't worry, that's not in itself bad.)
+#warning *****************************************************************
+#warning **** PyCUDA assumes that you are using the release (non-beta)
+#warning **** version. 3.0 beta will result in compile errors.
+#warning **** I apologize, but I have no way of checking whether I am
+#warning **** compiling against the beta--Nvidia cleverly changed the
+#warning **** API without bumping the version number.
+#warning **** You may try undefining CUDAPP_POST_30_BETA in 
+#warning **** src/cpp/cuda.hpp to get this to work.
+#warning *****************************************************************
+#endif
 
 
 
 
 // #define CUDAPP_TRACE_CUDA
+#define CUDAPP_POST_30_BETA
 
 
 
 
 #ifdef CUDAPP_TRACE_CUDA
-  #define CUDAPP_PRINT_CALL_TRACE(NAME) std::cerr << NAME << std::endl;
+  #define CUDAPP_PRINT_CALL_TRACE(NAME) \
+    std::cerr << NAME << std::endl;
+  #define CUDAPP_PRINT_CALL_TRACE_INFO(NAME, EXTRA_INFO) \
+    std::cerr << NAME << " (" << EXTRA_INFO << ')' << std::endl;
 #else
   #define CUDAPP_PRINT_CALL_TRACE(NAME) /*nothing*/
+  #define CUDAPP_PRINT_CALL_TRACE_INFO(NAME, EXTRA_INFO) /*nothing*/
 #endif
 
+#define CUDAPP_CALL_GUARDED_THREADED_WITH_TRACE_INFO(NAME, ARGLIST, TRACE_INFO) \
+  { \
+    CUDAPP_PRINT_CALL_TRACE_INFO(#NAME, TRACE_INFO); \
+    CUresult cu_status_code; \
+    Py_BEGIN_ALLOW_THREADS \
+      cu_status_code = NAME ARGLIST; \
+    Py_END_ALLOW_THREADS \
+    if (cu_status_code != CUDA_SUCCESS) \
+      throw cuda::error(#NAME, cu_status_code);\
+  }
+
+#define CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(NAME, ARGLIST, TRACE_INFO) \
+  { \
+    CUDAPP_PRINT_CALL_TRACE_INFO(#NAME, TRACE_INFO); \
+    CUresult cu_status_code; \
+    cu_status_code = NAME ARGLIST; \
+    if (cu_status_code != CUDA_SUCCESS) \
+      throw cuda::error(#NAME, cu_status_code);\
+  }
+
 #define CUDAPP_CALL_GUARDED_THREADED(NAME, ARGLIST) \
   { \
     CUDAPP_PRINT_CALL_TRACE(#NAME); \
@@ -53,6 +101,30 @@
     if (cu_status_code != CUDA_SUCCESS) \
       throw cuda::error(#NAME, cu_status_code);\
   }
+#define CUDAPP_CALL_GUARDED_CLEANUP(NAME, ARGLIST) \
+  { \
+    CUDAPP_PRINT_CALL_TRACE(#NAME); \
+    CUresult cu_status_code; \
+    cu_status_code = NAME ARGLIST; \
+    if (cu_status_code != CUDA_SUCCESS) \
+      std::cerr \
+        << "PyCUDA WARNING: a clean-up operation failed (dead context maybe?)" \
+        << std::endl \
+        << cuda::error::make_message(#NAME, cu_status_code) \
+        << std::endl; \
+  }
+#define CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(TYPE) \
+  catch (cuda::cannot_activate_out_of_thread_context) \
+  { } \
+  catch (cuda::cannot_activate_dead_context) \
+  { \
+    /* PyErr_Warn( \
+        PyExc_UserWarning, #TYPE " in dead context was implicitly cleaned up");*/ \
+  }
+  // In all likelihood, this TYPE's managing thread has exited, and
+  // therefore its context has already been deleted. No need to harp
+  // on the fact that we still thought there was cleanup to do.
+
 
 
 
@@ -69,8 +141,8 @@ namespace cuda
       const char *m_routine;
       CUresult m_code;
 
-    private:
-      static std::string make_message(const char *rout, CUresult c, const char *msg)
+    public:
+      static std::string make_message(const char *rout, CUresult c, const char *msg=0)
       {
         std::string result = rout;
         result += " failed: ";
@@ -83,7 +155,6 @@ namespace cuda
         return result;
       }
 
-    public:
       error(const char *rout, CUresult c, const char *msg=0)
         : std::runtime_error(make_message(rout, c, msg)),
         m_routine(rout), m_code(c)
@@ -124,6 +195,13 @@ namespace cuda
           case CUDA_ERROR_NO_BINARY_FOR_GPU: return "no binary for gpu";
           case CUDA_ERROR_ALREADY_ACQUIRED: return "already acquired";
           case CUDA_ERROR_NOT_MAPPED: return "not mapped";
+#if CUDA_VERSION >= 3000
+          case CUDA_ERROR_NOT_MAPPED_AS_ARRAY: return "not mapped as array";
+          case CUDA_ERROR_NOT_MAPPED_AS_POINTER: return "not mapped as pointer";
+#ifdef CUDAPP_POST_30_BETA
+          case CUDA_ERROR_ECC_UNCORRECTABLE: return "ECC uncorrectable";
+#endif
+#endif
 
           case CUDA_ERROR_INVALID_SOURCE: return "invalid source";
           case CUDA_ERROR_FILE_NOT_FOUND: return "file not found";
@@ -139,6 +217,11 @@ namespace cuda
           case CUDA_ERROR_LAUNCH_TIMEOUT: return "launch timeout";
           case CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING: return "launch incompatible texturing";
 
+#if CUDA_VERSION >= 3000
+          case CUDA_ERROR_POINTER_IS_64BIT: return "pointer is 64-bit";
+          case CUDA_ERROR_SIZE_IS_64BIT: return "size is 64-bit";
+#endif
+
           case CUDA_ERROR_UNKNOWN: return "unknown";
 
           default: return "invalid error code";
@@ -146,6 +229,32 @@ namespace cuda
       }
   };
 
+  struct cannot_activate_out_of_thread_context : public std::logic_error
+  { 
+    cannot_activate_out_of_thread_context(std::string const &w)
+      : std::logic_error(w)
+    { }
+  };
+
+  struct cannot_activate_dead_context : public std::logic_error
+  { 
+    cannot_activate_dead_context(std::string const &w)
+      : std::logic_error(w)
+    { }
+  };
+
+
+
+  // version query ------------------------------------------------------------
+#if CUDA_VERSION >= 2020
+  inline int get_driver_version()
+  {
+    int result;
+    CUDAPP_CALL_GUARDED(cuDriverGetVersion, (&result));
+    return result;
+  }
+#endif
+
 
 
 
@@ -190,7 +299,7 @@ namespace cuda
         return bytes;
       }
 
-      int get_attribute(CUdevice_attribute attr)
+      int get_attribute(CUdevice_attribute attr) const
       {
         int result;
         CUDAPP_CALL_GUARDED(cuDeviceGetAttribute, (&result, attr, m_device));
@@ -213,6 +322,10 @@ namespace cuda
       }
 
       boost::shared_ptr<context> make_context(unsigned int flags);
+
+      CUdevice handle() const
+      { return m_device; }
+
   };
 
   inline
@@ -238,66 +351,141 @@ namespace cuda
    * to push contexts that are already active at a deeper stack level, so we
    * maintain all contexts floating other than the top one.
    */
+
+  // for friend decl
+  namespace gl { 
+    boost::shared_ptr<context> 
+        make_gl_context(device const &dev, unsigned int flags);
+  }
+
+  class context_stack;
+  extern boost::thread_specific_ptr<context_stack> context_stack_ptr;
+
+  class context_stack
+  {
+      /* This wrapper is necessary because we need to pop the contents
+       * off the stack before we destroy each of the contexts. This, in turn,
+       * is because the contexts need to be able to access the stack in order
+       * to be destroyed.
+       */
+    private:
+      typedef std::stack<boost::shared_ptr<context> > stack_t;
+      typedef stack_t::value_type value_type;;
+      stack_t m_stack;
+
+    public:
+      ~context_stack();
+
+      bool empty() const
+      { return m_stack.empty(); }
+
+      value_type &top()
+      { return m_stack.top(); }
+
+      void pop()
+      { m_stack.pop(); }
+
+      void push(value_type v)
+      { m_stack.push(v); }
+
+      static context_stack &get()
+      {
+        if (context_stack_ptr.get() == 0)
+          context_stack_ptr.reset(new context_stack);
+
+        return *context_stack_ptr;
+      }
+  };
+
   class context : boost::noncopyable
   {
     private:
       CUcontext m_context;
       bool m_valid;
       unsigned m_use_count;
-      typedef std::stack<boost::weak_ptr<context>,
-        std::vector<boost::weak_ptr<context> > > context_stack_t;
-      
-      static context_stack_t m_context_stack;
+      boost::thread::id m_thread;
 
     public:
       context(CUcontext ctx)
-        : m_context(ctx), m_valid(true), m_use_count(1)
+        : m_context(ctx), m_valid(true), m_use_count(1), 
+        m_thread(boost::this_thread::get_id())
       { }
 
-
       ~context()
-      { 
+      {
         if (m_valid)
         {
-          if (m_use_count)
-            std::cerr 
-              << "-----------------------------------------------------------" << std::endl
-              << "PyCuda WARNING: I'm being asked to destroy a " << std::endl
-              << "context that's part of the current context stack." << std::endl
-              << "-----------------------------------------------------------" << std::endl
-              << "I will pick the next lower active context from the" << std::endl
-              << "context stack. Since this choice is happening" << std::endl
-              << "at an unspecified point in time, your code" << std::endl
-              << "may be making false assumptions about which" << std::endl
-              << "context is active at what point." << std::endl
-              << "Call Context.pop() to avoid this warning." << std::endl
-              << "-----------------------------------------------------------" << std::endl
-              << "If Python is terminating abnormally (eg. exiting upon an" << std::endl
-              << "unhandled exception), you may ignore this." << std::endl
-              << "-----------------------------------------------------------" << std::endl;
+          /* It's possible that we get here with a non-zero m_use_count. Since the context
+           * stack holds shared_ptrs, this must mean that the context stack itself is getting
+           * destroyed, which means it's ok for this context to sign off, too.
+           */
           detach();
         }
       }
 
-      CUcontext data() const
+      CUcontext handle() const
       { return m_context; }
 
+      bool operator==(const context &other) const
+      {
+        return m_context == other.m_context;
+      }
+
+      bool operator!=(const context &other) const
+      {
+        return m_context != other.m_context;
+      }
+
+      long hash() const
+      {
+        return long(m_context) ^ long(this);
+      }
+
+      boost::thread::id thread_id() const
+      { return m_thread; }
+
+      bool is_valid() const
+      {
+        return m_valid;
+      }
+
       void detach()
       {
         if (m_valid)
         {
-          if (current_context().get() == this)
+          bool active_before_destruction = current_context().get() == this;
+          if (active_before_destruction)
           {
-            CUDAPP_CALL_GUARDED(cuCtxDetach, (m_context));
+            CUDAPP_CALL_GUARDED_CLEANUP(cuCtxDetach, (m_context));
           }
           else
-            CUDAPP_CALL_GUARDED(cuCtxDestroy, (m_context));
+          {
+            if (m_thread == boost::this_thread::get_id())
+            {
+              CUDAPP_CALL_GUARDED_CLEANUP(cuCtxPushCurrent, (m_context));
+              CUDAPP_CALL_GUARDED_CLEANUP(cuCtxDetach, (m_context));
+              /* pop is implicit in detach */
+            }
+            else
+            {
+              // In all likelihood, this context's managing thread has exited, and 
+              // therefore this context has already been deleted. No need to harp
+              // on the fact that we still thought there was cleanup to do.
+
+              // std::cerr << "PyCUDA WARNING: leaked out-of-thread context " << std::endl;
+            }
+          }
 
           m_valid = false;
 
-          boost::shared_ptr<context> new_active = current_context(this);
-          if (new_active.get())
-            CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (new_active->m_context)); 
+          if (active_before_destruction)
+          {
+            boost::shared_ptr<context> new_active = current_context(this);
+            if (new_active.get())
+            {
+              CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (new_active->m_context)); 
+            }
+          }
         }
         else
           throw error("context::detach", CUDA_ERROR_INVALID_CONTEXT,
@@ -315,20 +503,31 @@ namespace cuda
 
       static void prepare_context_switch()
       {
-        if (context::m_context_stack.size())
+        if (!context_stack::get().empty())
         {
           CUcontext popped;
           CUDAPP_CALL_GUARDED(cuCtxPopCurrent, (&popped)); 
         }
       }
 
-      void pop()
+      static void pop()
       { 
         prepare_context_switch();
-        m_context_stack.pop();
-        --m_use_count;
+        context_stack &ctx_stack = context_stack::get();
+
+        if (ctx_stack.empty())
+        {
+          throw error("context::pop", CUDA_ERROR_INVALID_CONTEXT,
+              "cannot pop non-current context");
+        }
 
         boost::shared_ptr<context> current = current_context();
+        if (current)
+          --current->m_use_count;
+
+        ctx_stack.pop();
+
+        current = current_context();
         if (current)
           CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (current_context()->m_context)); 
       }
@@ -343,25 +542,26 @@ namespace cuda
       {
         while (true)
         {
-          if (m_context_stack.size() == 0)
+          if (context_stack::get().empty())
             return boost::shared_ptr<context>();
-          boost::weak_ptr<context> result(m_context_stack.top());
-          if (!result.expired() && result.lock().get() != except)
+
+          boost::shared_ptr<context> result(context_stack::get().top());
+          if (result.get() != except 
+              && result->is_valid())
           {
             // good, weak pointer didn't expire
-            // (treating except as expired weak pointer)
-            return result.lock();
-          }
-          else
-          {
-            // weak pointer invalidated, pop it and try again.
-            m_context_stack.pop();
+            return result;
           }
+
+          // context invalid, pop it and try again.
+          context_stack::get().pop();
         }
       }
 
       friend class device;
       friend void context_push(boost::shared_ptr<context> ctx);
+      friend boost::shared_ptr<context> 
+          gl::make_gl_context(device const &dev, unsigned int flags);
   };
 
 
@@ -375,7 +575,7 @@ namespace cuda
     CUcontext ctx;
     CUDAPP_CALL_GUARDED(cuCtxCreate, (&ctx, flags, m_device));
     boost::shared_ptr<context> result(new context(ctx));
-    context::m_context_stack.push(result);
+    context_stack::get().push(result);
     return result;
   }
 
@@ -389,7 +589,7 @@ namespace cuda
     context::prepare_context_switch();
 
     CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (ctx->m_context)); 
-    context::m_context_stack.push(ctx);
+    context_stack::get().push(ctx);
     ++ctx->m_use_count;
   }
 #endif
@@ -397,6 +597,27 @@ namespace cuda
 
 
 
+  inline context_stack::~context_stack()
+  {
+    if (!m_stack.empty())
+    {
+      std::cerr 
+        << "-------------------------------------------------------------------" << std::endl
+        << "PyCUDA ERROR: The context stack was not empty upon module cleanup." << std::endl
+        << "-------------------------------------------------------------------" << std::endl
+        << "A context was still active when the context stack was being" << std::endl
+        << "cleaned up. At this point in our execution, CUDA may already" << std::endl
+        << "have been deinitialized, so there is no way we can finish" << std::endl
+        << "cleanly. The program will be aborted now." << std::endl
+        << "Use Context.pop() to avoid this problem." << std::endl
+        << "-------------------------------------------------------------------" << std::endl;
+      abort();
+    }
+  }
+
+
+
+
   class explicit_context_dependent
   {
     private:
@@ -444,9 +665,16 @@ namespace cuda
       scoped_context_activation(boost::shared_ptr<context> ctx)
         : m_context(ctx)
       { 
+        if (!m_context->is_valid())
+          throw cuda::cannot_activate_dead_context(
+              "cannot activate dead context");
+
         m_did_switch = context::current_context() != m_context;
         if (m_did_switch)
         {
+          if (boost::this_thread::get_id() != m_context->thread_id())
+            throw cuda::cannot_activate_out_of_thread_context(
+                "cannot activate out-of-thread context");
 #if CUDA_VERSION >= 2000
           context_push(m_context);
 #else
@@ -478,14 +706,18 @@ namespace cuda
 
       ~stream()
       { 
-        scoped_context_activation ca(get_context());
-        CUDAPP_CALL_GUARDED(cuStreamDestroy, (m_stream)); 
+        try
+        {
+          scoped_context_activation ca(get_context());
+          CUDAPP_CALL_GUARDED_CLEANUP(cuStreamDestroy, (m_stream)); 
+        }
+        CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(stream);
       }
 
       void synchronize()
       { CUDAPP_CALL_GUARDED_THREADED(cuStreamSynchronize, (m_stream)); }
 
-      CUstream data() const
+      CUstream handle() const
       { return m_stream; }
 
       bool is_done() const
@@ -538,10 +770,12 @@ namespace cuda
       {
         if (m_managed)
         {
+          try
           {
             scoped_context_activation ca(get_context());
-            CUDAPP_CALL_GUARDED(cuArrayDestroy, (m_array)); 
+            CUDAPP_CALL_GUARDED_CLEANUP(cuArrayDestroy, (m_array)); 
           }
+          CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(array);
 
           m_managed = false;
           release_context();
@@ -564,7 +798,7 @@ namespace cuda
       }
 #endif
 
-      CUarray data() const
+      CUarray handle() const
       { return m_array; }
   };
 
@@ -597,32 +831,45 @@ namespace cuda
       { 
         if (m_managed)
         {
-          CUDAPP_CALL_GUARDED(cuTexRefDestroy, (m_texref)); 
+          CUDAPP_CALL_GUARDED_CLEANUP(cuTexRefDestroy, (m_texref)); 
         }
       }
 
       void set_module(boost::shared_ptr<module> mod)
       { m_module = mod; }
 
-      CUtexref data() const
+      CUtexref handle() const
       { return m_texref; }
 
       void set_array(boost::shared_ptr<array> ary)
       { 
         CUDAPP_CALL_GUARDED(cuTexRefSetArray, (m_texref, 
-            ary->data(), CU_TRSA_OVERRIDE_FORMAT)); 
+            ary->handle(), CU_TRSA_OVERRIDE_FORMAT)); 
         m_array = ary;
       }
 
-      unsigned int set_address(CUdeviceptr dptr, unsigned int bytes)
+      unsigned int set_address(CUdeviceptr dptr, unsigned int bytes, bool allow_offset=false)
       { 
         unsigned int byte_offset;
         CUDAPP_CALL_GUARDED(cuTexRefSetAddress, (&byte_offset,
               m_texref, dptr, bytes)); 
+
+        if (!allow_offset && byte_offset != 0)
+          throw cuda::error("texture_reference::set_address", CUDA_ERROR_INVALID_VALUE,
+              "texture binding resulted in offset, but allow_offset was false");
+
         m_array.reset();
         return byte_offset;
       }
 
+#if CUDA_VERSION >= 2020
+      void set_address_2d(CUdeviceptr dptr, 
+          const CUDA_ARRAY_DESCRIPTOR &descr, unsigned int pitch)
+      { 
+        CUDAPP_CALL_GUARDED(cuTexRefSetAddress2D, (m_texref, &descr, dptr, pitch)); 
+      }
+#endif
+
       void set_format(CUarray_format fmt, int num_packed_components)
       { CUDAPP_CALL_GUARDED(cuTexRefSetFormat, (m_texref, fmt, num_packed_components)); }
 
@@ -695,11 +942,15 @@ namespace cuda
 
       ~module()
       {
-        scoped_context_activation ca(get_context());
-        CUDAPP_CALL_GUARDED(cuModuleUnload, (m_module));
+        try
+        {
+          scoped_context_activation ca(get_context());
+          CUDAPP_CALL_GUARDED_CLEANUP(cuModuleUnload, (m_module));
+        }
+        CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(module);
       }
 
-      CUmodule data() const
+      CUmodule handle() const
       { return m_module; }
 
       function get_function(const char *name);
@@ -724,7 +975,7 @@ namespace cuda
   texture_reference *module_get_texref(boost::shared_ptr<module> mod, const char *name)
   {
     CUtexref tr;
-    CUDAPP_CALL_GUARDED(cuModuleGetTexRef, (&tr, mod->data(), name));
+    CUDAPP_CALL_GUARDED(cuModuleGetTexRef, (&tr, mod->handle(), name));
     std::auto_ptr<texture_reference> result(
         new texture_reference(tr, false));
     result->set_module(mod);
@@ -739,39 +990,84 @@ namespace cuda
   {
     private:
       CUfunction m_function;
+      std::string m_symbol;
 
     public:
-      function(CUfunction func)
-        : m_function(func)
+      function(CUfunction func, std::string const &sym)
+        : m_function(func), m_symbol(sym)
       { }
 
       void set_block_shape(int x, int y, int z)
-      { CUDAPP_CALL_GUARDED(cuFuncSetBlockShape, (m_function, x, y, z)); }
+      { 
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
+            cuFuncSetBlockShape, (m_function, x, y, z), m_symbol); 
+      }
       void set_shared_size(unsigned int bytes)
-      { CUDAPP_CALL_GUARDED(cuFuncSetSharedSize, (m_function, bytes)); }
+      { 
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
+            cuFuncSetSharedSize, (m_function, bytes), m_symbol); 
+      }
 
       void param_set_size(unsigned int bytes)
-      { CUDAPP_CALL_GUARDED(cuParamSetSize, (m_function, bytes)); }
+      { 
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
+            cuParamSetSize, (m_function, bytes), m_symbol); 
+      }
       void param_set(int offset, unsigned int value)
-      { CUDAPP_CALL_GUARDED(cuParamSeti, (m_function, offset, value)); }
+      { 
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
+            cuParamSeti, (m_function, offset, value), m_symbol); 
+      }
       void param_set(int offset, float value)
-      { CUDAPP_CALL_GUARDED(cuParamSetf, (m_function, offset, value)); }
+      { 
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
+          cuParamSetf, (m_function, offset, value), m_symbol); 
+      }
       void param_setv(int offset, void *buf, unsigned long len)
       { 
-        CUDAPP_CALL_GUARDED(cuParamSetv, (m_function, offset, buf, len)); 
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
+            cuParamSetv, (m_function, offset, buf, len), m_symbol); 
       }
       void param_set_texref(const texture_reference &tr)
       { 
-        CUDAPP_CALL_GUARDED(cuParamSetTexRef, (m_function, 
-            CU_PARAM_TR_DEFAULT, tr.data())); 
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(cuParamSetTexRef, (m_function, 
+            CU_PARAM_TR_DEFAULT, tr.handle()), m_symbol); 
       }
 
       void launch()
-      { CUDAPP_CALL_GUARDED_THREADED(cuLaunch, (m_function)); }
+      { 
+        CUDAPP_CALL_GUARDED_THREADED_WITH_TRACE_INFO(
+            cuLaunch, (m_function), m_symbol); 
+      }
       void launch_grid(int grid_width, int grid_height)
-      { CUDAPP_CALL_GUARDED_THREADED(cuLaunchGrid, (m_function, grid_width, grid_height)); }
+      { 
+        CUDAPP_CALL_GUARDED_THREADED_WITH_TRACE_INFO(
+          cuLaunchGrid, (m_function, grid_width, grid_height), m_symbol); 
+      }
       void launch_grid_async(int grid_width, int grid_height, const stream &s)
-      { CUDAPP_CALL_GUARDED_THREADED(cuLaunchGridAsync, (m_function, grid_width, grid_height, s.data())); }
+      { 
+        CUDAPP_CALL_GUARDED_THREADED_WITH_TRACE_INFO(
+            cuLaunchGridAsync, (m_function, grid_width, grid_height, s.handle()), 
+            m_symbol);
+      }
+
+#if CUDA_VERSION >= 2020
+      int get_attribute(CUfunction_attribute attr) const
+      {
+        int result;
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
+            cuFuncGetAttribute, (&result, attr, m_function), m_symbol);
+        return result;
+      }
+#endif
+
+#if CUDA_VERSION >= 3000 && defined(CUDAPP_POST_30_BETA)
+      void set_cache_config(CUfunc_cache fc)
+      {
+        CUDAPP_CALL_GUARDED_WITH_TRACE_INFO(
+            cuFuncSetCacheConfig, (m_function, fc), m_symbol);
+      }
+#endif
   };
 
   inline
@@ -779,7 +1075,7 @@ namespace cuda
   {
     CUfunction func;
     CUDAPP_CALL_GUARDED(cuModuleGetFunction, (&func, m_module, name));
-    return function(func);
+    return function(func, name);
   }
 
 
@@ -805,7 +1101,7 @@ namespace cuda
   inline 
   void mem_free(CUdeviceptr devptr)
   {
-    CUDAPP_CALL_GUARDED(cuMemFree, (devptr));
+    CUDAPP_CALL_GUARDED_CLEANUP(cuMemFree, (devptr));
   }
 
   class device_allocation : public boost::noncopyable, public context_dependent
@@ -825,8 +1121,13 @@ namespace cuda
       {
         if (m_valid)
         {
-          scoped_context_activation ca(get_context());
-          mem_free(m_devptr);
+          try
+          {
+            scoped_context_activation ca(get_context());
+            mem_free(m_devptr);
+          }
+          CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(device_allocation);
+
           release_context();
           m_valid = false;
         }
@@ -868,18 +1169,18 @@ namespace cuda
 
   inline
   void memcpy_dtoa(array const &ary, unsigned int index, CUdeviceptr src, unsigned int len)
-  { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyDtoA, (ary.data(), index, src, len)); }
+  { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyDtoA, (ary.handle(), index, src, len)); }
 
   inline
   void memcpy_atod(CUdeviceptr dst, array const &ary, unsigned int index, unsigned int len)
-  { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyAtoD, (dst, ary.data(), index, len)); }
+  { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyAtoD, (dst, ary.handle(), index, len)); }
 
   inline
   void memcpy_atoa(
       array const &dst, unsigned int dst_index, 
       array const &src, unsigned int src_index, 
       unsigned int len)
-  { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyAtoA, (dst.data(), dst_index, src.data(), src_index, len)); }
+  { CUDAPP_CALL_GUARDED_THREADED(cuMemcpyAtoA, (dst.handle(), dst_index, src.handle(), src_index, len)); }
 
 
 
@@ -903,7 +1204,7 @@ namespace cuda
     void set_src_array(array const &ary)  \
     {  \
       srcMemoryType = CU_MEMORYTYPE_ARRAY; \
-      srcArray = ary.data();  \
+      srcArray = ary.handle();  \
     } \
     \
     void set_src_device(CUdeviceptr devptr)  \
@@ -923,7 +1224,7 @@ namespace cuda
     void set_dst_array(array const &ary) \
     { \
       dstMemoryType = CU_MEMORYTYPE_ARRAY; \
-      dstArray = ary.data(); \
+      dstArray = ary.handle(); \
     } \
     \
     void set_dst_device(CUdeviceptr devptr)  \
@@ -958,7 +1259,7 @@ namespace cuda
     }
 
     void execute_async(const stream &s) const
-    { CUDAPP_CALL_GUARDED_THREADED(cuMemcpy2DAsync, (this, s.data())); }
+    { CUDAPP_CALL_GUARDED_THREADED(cuMemcpy2DAsync, (this, s.handle())); }
   };
 
 #if CUDA_VERSION >= 2000
@@ -988,51 +1289,81 @@ namespace cuda
     }
 
     void execute_async(const stream &s) const
-    { CUDAPP_CALL_GUARDED_THREADED(cuMemcpy3DAsync, (this, s.data())); }
+    { CUDAPP_CALL_GUARDED_THREADED(cuMemcpy3DAsync, (this, s.handle())); }
   };
 #endif
 
 
   // host memory --------------------------------------------------------------
-  inline void *mem_alloc_host(unsigned int size)
+  inline void *mem_alloc_host(unsigned int size, unsigned flags=0)
   {
     void *m_data;
+#if CUDA_VERSION >= 2020
+    CUDAPP_CALL_GUARDED(cuMemHostAlloc, (&m_data, size, flags));
+#else
+    if (flags != 0)
+      throw cuda::error("mem_alloc_host", CUDA_ERROR_INVALID_VALUE,
+          "nonzero flags in mem_alloc_host not allowed in CUDA 2.1 and older");
     CUDAPP_CALL_GUARDED(cuMemAllocHost, (&m_data, size));
+#endif
     return m_data;
   }
 
   inline void mem_free_host(void *ptr)
   {
-    CUDAPP_CALL_GUARDED(cuMemFreeHost, (ptr));
+    CUDAPP_CALL_GUARDED_CLEANUP(cuMemFreeHost, (ptr));
   }
 
 
 
 
-  struct host_allocation : public boost::noncopyable
+  struct host_allocation : public boost::noncopyable, public context_dependent
   {
     private:
+      bool m_valid;
       void *m_data;
 
     public:
-      host_allocation(unsigned bytesize)
-        : m_data(mem_alloc_host(bytesize))
+      host_allocation(unsigned bytesize, unsigned flags=0)
+        : m_valid(true), m_data(mem_alloc_host(bytesize, flags))
       { }
 
       ~host_allocation()
-      { free(); }
+      { 
+        if (m_valid)
+          free(); 
+      }
 
       void free()
       {
-        if (m_data)
+        if (m_valid)
         {
-          mem_free_host(m_data); 
-          m_data = 0;
+          try
+          {
+            scoped_context_activation ca(get_context());
+            mem_free_host(m_data); 
+          }
+          CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(host_allocation);
+
+          release_context();
+          m_valid = false;
         }
+        else
+          throw cuda::error("host_allocation::free", CUDA_ERROR_INVALID_HANDLE);
       }
       
       void *data()
       { return m_data; }
+
+#if CUDA_VERSION >= 2020
+      CUdeviceptr get_device_pointer()
+      {
+        CUdeviceptr result;
+        CUDAPP_CALL_GUARDED(cuMemHostGetDevicePointer, (&result, m_data, 0));
+        return result;
+      }
+#endif
+
   };
 
 
@@ -1050,18 +1381,34 @@ namespace cuda
 
       ~event()
       { 
-        scoped_context_activation ca(get_context());
-        CUDAPP_CALL_GUARDED(cuEventDestroy, (m_event)); 
+        try
+        {
+          scoped_context_activation ca(get_context());
+          CUDAPP_CALL_GUARDED_CLEANUP(cuEventDestroy, (m_event)); 
+        }
+        CUDAPP_CATCH_CLEANUP_ON_DEAD_CONTEXT(event);
       }
 
-      void record()
-      { CUDAPP_CALL_GUARDED(cuEventRecord, (m_event, 0)); }
+      event *record(py::object stream_py)
+      { 
+        CUstream s_handle;
+        if (stream_py.ptr() != Py_None)
+        {
+          const stream &s = py::extract<const stream &>(stream_py);
+          s_handle = s.handle();
+        }
+        else
+          s_handle = 0;
 
-      void record_in_stream(stream const &str)
-      { CUDAPP_CALL_GUARDED(cuEventRecord, (m_event, str.data())); }
+        CUDAPP_CALL_GUARDED(cuEventRecord, (m_event, s_handle)); 
+        return this;
+      }
 
-      void synchronize()
-      { CUDAPP_CALL_GUARDED_THREADED(cuEventSynchronize, (m_event)); }
+      event *synchronize()
+      { 
+        CUDAPP_CALL_GUARDED_THREADED(cuEventSynchronize, (m_event)); 
+        return this;
+      }
 
       bool query() const
       { 
