@@ -1,29 +1,39 @@
 from __future__ import division
-import unittest
-import pycuda.autoinit
-import pycuda.driver as drv
 import numpy
 import numpy.linalg as la
+from pycuda.tools import mark_cuda_test
 
 
 
 
-assert isinstance(pycuda.autoinit.device.name(), str)
-assert isinstance(pycuda.autoinit.device.compute_capability(), tuple)
-assert isinstance(pycuda.autoinit.device.get_attributes(), dict)
+def have_pycuda():
+    try:
+        import pycuda
+        return True
+    except:
+        return False
+
+
+if have_pycuda():
+    import pycuda.gpuarray as gpuarray
+    import pycuda.driver as drv
+    from pycuda.compiler import SourceModule
 
 
 
 
+class TestDriver:
+    disabled = not have_pycuda()
 
-class TestCuda(unittest.TestCase):
+    @mark_cuda_test
     def test_memory(self):
         z = numpy.random.randn(400).astype(numpy.float32)
         new_z = drv.from_device_like(drv.to_device(z), z)
         assert la.norm(new_z-z) == 0
 
+    @mark_cuda_test
     def test_simple_kernel(self):
-        mod = drv.SourceModule("""
+        mod = SourceModule("""
         __global__ void multiply_them(float *dest, float *a, float *b)
         {
           const int i = threadIdx.x;
@@ -41,10 +51,11 @@ class TestCuda(unittest.TestCase):
         multiply_them(
                 drv.Out(dest), drv.In(a), drv.In(b),
                 block=(400,1,1))
-        self.assert_(la.norm(dest-a*b) == 0)
+        assert la.norm(dest-a*b) == 0
 
+    @mark_cuda_test
     def test_simple_kernel_2(self):
-        mod = drv.SourceModule("""
+        mod = SourceModule("""
         __global__ void multiply_them(float *dest, float *a, float *b)
         {
           const int i = threadIdx.x;
@@ -64,7 +75,7 @@ class TestCuda(unittest.TestCase):
         multiply_them(
                 drv.Out(dest), a_gpu, b_gpu,
                 block=(400,1,1))
-        self.assert_(la.norm(dest-a*b) == 0)
+        assert la.norm(dest-a*b) == 0
 
         # now try with offsets
         dest = numpy.zeros_like(a)
@@ -72,14 +83,15 @@ class TestCuda(unittest.TestCase):
                 drv.Out(dest), numpy.intp(a_gpu)+1, b_gpu,
                 block=(399,1,1))
 
-        self.assert_(la.norm(dest-a*b) == 0)
+        assert la.norm(dest-a*b) == 0
 
+    @mark_cuda_test
     def test_streamed_kernel(self):
         # this differs from the "simple_kernel" case in that *all* computation
         # and data copying is asynchronous. Observe how this necessitates the
         # use of page-locked memory.
 
-        mod = drv.SourceModule("""
+        mod = SourceModule("""
         __global__ void multiply_them(float *dest, float *a, float *b)
         {
           const int i = threadIdx.x*blockDim.y + threadIdx.y;
@@ -96,19 +108,30 @@ class TestCuda(unittest.TestCase):
         a[:] = numpy.random.randn(*shape)
         b[:] = numpy.random.randn(*shape)
 
+        a_gpu = drv.mem_alloc(a.nbytes)
+        b_gpu = drv.mem_alloc(b.nbytes)
+
         strm = drv.Stream()
+        drv.memcpy_htod_async(a_gpu, a, strm)
+        drv.memcpy_htod_async(b_gpu, b, strm)
+        strm.synchronize()
 
         dest = drv.pagelocked_empty_like(a)
         multiply_them(
-                drv.Out(dest), drv.In(a), drv.In(b),
+                drv.Out(dest), a_gpu, b_gpu,
                 block=shape+(1,), stream=strm)
         strm.synchronize()
 
-        self.assert_(la.norm(dest-a*b) == 0)
+        drv.memcpy_dtoh_async(a, a_gpu, strm)
+        drv.memcpy_dtoh_async(b, b_gpu, strm)
+        strm.synchronize()
 
+        assert la.norm(dest-a*b) == 0
+
+    @mark_cuda_test
     def test_gpuarray(self):
         import numpy
-        a = numpy.arange(2000000, dtype=numpy.float32)
+        a = numpy.arange(200000, dtype=numpy.float32)
         b = a + 17
         import pycuda.gpuarray as gpuarray
         a_g = gpuarray.to_gpu(a)
@@ -119,19 +142,22 @@ class TestCuda(unittest.TestCase):
         diff = ((a_g*b_g).get()-a*b)
         assert la.norm(diff) == 0
 
-    def donottest_cublas_mixing(self):
-        self.test_streamed_kernel()
+    @mark_cuda_test
+    def donottest_cublas_mixing():
+        test_streamed_kernel()
 
         import pycuda.blas as blas
 
         shape = (10,)
         a = blas.ones(shape, dtype=numpy.float32)
         b = 33*blas.ones(shape, dtype=numpy.float32)
-        self.assert_(((-a+b).from_gpu() == 32).all())
-        self.test_streamed_kernel()
+        assert ((-a+b).from_gpu() == 32).all()
 
+        test_streamed_kernel()
+
+    @mark_cuda_test
     def test_2d_texture(self):
-        mod = drv.SourceModule("""
+        mod = SourceModule("""
         texture<float, 2, cudaReadModeElementType> mtx_tex;
 
         __global__ void copy_texture(float *dest)
@@ -152,13 +178,14 @@ class TestCuda(unittest.TestCase):
 
         dest = numpy.zeros(shape, dtype=numpy.float32)
         copy_texture(drv.Out(dest),
-                block=shape+(1,), 
+                block=shape+(1,),
                 texrefs=[mtx_tex]
                 )
         assert la.norm(dest-a) == 0
 
+    @mark_cuda_test
     def test_multiple_2d_textures(self):
-        mod = drv.SourceModule("""
+        mod = SourceModule("""
         texture<float, 2, cudaReadModeElementType> mtx_tex;
         texture<float, 2, cudaReadModeElementType> mtx2_tex;
 
@@ -167,7 +194,7 @@ class TestCuda(unittest.TestCase):
           int row = threadIdx.x;
           int col = threadIdx.y;
           int w = blockDim.y;
-          dest[row*w+col] = 
+          dest[row*w+col] =
               tex2D(mtx_tex, row, col)
               +
               tex2D(mtx2_tex, row, col);
@@ -186,13 +213,14 @@ class TestCuda(unittest.TestCase):
 
         dest = numpy.zeros(shape, dtype=numpy.float32)
         copy_texture(drv.Out(dest),
-                block=shape+(1,), 
+                block=shape+(1,),
                 texrefs=[mtx_tex, mtx2_tex]
                 )
         assert la.norm(dest-a-b) < 1e-6
 
+    @mark_cuda_test
     def test_multichannel_2d_texture(self):
-        mod = drv.SourceModule("""
+        mod = SourceModule("""
         #define CHANNELS 4
         texture<float4, 2, cudaReadModeElementType> mtx_tex;
 
@@ -214,13 +242,15 @@ class TestCuda(unittest.TestCase):
 
         shape = (5,6)
         channels = 4
-        a = numpy.random.randn(*((channels,)+shape)).astype(numpy.float32)
+        a = numpy.asarray(
+                numpy.random.randn(*((channels,)+shape)),
+                dtype=numpy.float32, order="F")
         drv.bind_array_to_texref(
             drv.make_multichannel_2d_array(a, order="F"), mtx_tex)
 
         dest = numpy.zeros(shape+(channels,), dtype=numpy.float32)
         copy_texture(drv.Out(dest),
-                block=shape+(1,), 
+                block=shape+(1,),
                 texrefs=[mtx_tex]
                 )
         reshaped_a = a.transpose(1,2,0)
@@ -228,9 +258,46 @@ class TestCuda(unittest.TestCase):
         #print dest
         assert la.norm(dest-reshaped_a) == 0
 
+    @mark_cuda_test
+    def test_multichannel_linear_texture(self):
+        mod = SourceModule("""
+        #define CHANNELS 4
+        texture<float4, 1, cudaReadModeElementType> mtx_tex;
+
+        __global__ void copy_texture(float *dest)
+        {
+          int i = threadIdx.x+blockDim.x*threadIdx.y;
+          float4 texval = tex1Dfetch(mtx_tex, i);
+          dest[i*CHANNELS + 0] = texval.x;
+          dest[i*CHANNELS + 1] = texval.y;
+          dest[i*CHANNELS + 2] = texval.z;
+          dest[i*CHANNELS + 3] = texval.w;
+        }
+        """)
+
+        copy_texture = mod.get_function("copy_texture")
+        mtx_tex = mod.get_texref("mtx_tex")
+
+        shape = (16, 16)
+        channels = 4
+        a = numpy.random.randn(*(shape+(channels,))).astype(numpy.float32)
+        a_gpu = drv.to_device(a)
+        mtx_tex.set_address(a_gpu, a.nbytes)
+        mtx_tex.set_format(drv.array_format.FLOAT, 4)
+
+        dest = numpy.zeros(shape+(channels,), dtype=numpy.float32)
+        copy_texture(drv.Out(dest),
+                block=shape+(1,),
+                texrefs=[mtx_tex]
+                )
+        #print a
+        #print dest
+        assert la.norm(dest-a) == 0
+
+    @mark_cuda_test
     def test_large_smem(self):
         n = 4000
-        mod = drv.SourceModule("""
+        mod = SourceModule("""
         #include <stdio.h>
 
         __global__ void kernel(int *d_data)
@@ -248,6 +315,7 @@ class TestCuda(unittest.TestCase):
 
         kernel(arg, block=(1,1,1,), )
 
+    @mark_cuda_test
     def test_bitlog(self):
         from pycuda.tools import bitlog2
         assert bitlog2(17) == 4
@@ -255,6 +323,7 @@ class TestCuda(unittest.TestCase):
         assert bitlog2(0x3affe) == 17
         assert bitlog2(0xcc3affe) == 27
 
+    @mark_cuda_test
     def test_mempool_2(self):
         from pycuda.tools import DeviceMemoryPool as DMP
         from random import randrange
@@ -267,7 +336,8 @@ class TestCuda(unittest.TestCase):
             assert asize >= s, s
             assert DMP.bin_number(asize) == bin_nr, s
             assert asize < asize*(1+1/8)
-            
+
+    @mark_cuda_test
     def test_mempool(self):
         from pycuda.tools import bitlog2
         from pycuda.tools import DeviceMemoryPool
@@ -279,24 +349,31 @@ class TestCuda(unittest.TestCase):
 
         e0 = bitlog2(free)
 
-        for e in range(e0-5, e0-3):
+        for e in range(e0-6, e0-4):
             for i in range(100):
                 queue.append(pool.allocate(1<<e))
                 if len(queue) > 10:
                     queue.pop(0)
+        del queue
+        pool.stop_holding()
 
+    @mark_cuda_test
     def test_multi_context(self):
         if drv.get_version() < (2,0,0):
             return
+        if drv.get_version() >= (2,2,0):
+            if drv.Context.get_device().compute_mode == drv.compute_mode.EXCLUSIVE:
+                return
 
         mem_a = drv.mem_alloc(50)
-        ctx2 = pycuda.autoinit.device.make_context()
+        ctx2 = drv.Context.get_device().make_context()
         mem_b = drv.mem_alloc(60)
 
         del mem_a
         del mem_b
         ctx2.detach()
 
+    @mark_cuda_test
     def test_3d_texture(self):
         # adapted from code by Nicolas Pinto
         w = 2
@@ -327,7 +404,7 @@ class TestCuda(unittest.TestCase):
 
         copy()
 
-        mod = drv.SourceModule("""
+        mod = SourceModule("""
         texture<float, 3, cudaReadModeElementType> mtx_tex;
 
         __global__ void copy_texture(float *dest)
@@ -352,13 +429,14 @@ class TestCuda(unittest.TestCase):
         copy_texture(drv.Out(dest), block=shape, texrefs=[mtx_tex])
         assert la.norm(dest-a) == 0
 
+    @mark_cuda_test
     def test_prepared_invocation(self):
         a = numpy.random.randn(4,4).astype(numpy.float32)
         a_gpu = drv.mem_alloc(a.size * a.dtype.itemsize)
 
         drv.memcpy_htod(a_gpu, a)
 
-        mod = drv.SourceModule("""
+        mod = SourceModule("""
             __global__ void doublify(float *a)
             {
               int idx = threadIdx.x + threadIdx.y*blockDim.x;
@@ -380,7 +458,83 @@ class TestCuda(unittest.TestCase):
         drv.memcpy_dtoh(a_quadrupled, a_gpu)
         assert la.norm(a_quadrupled[1:]-4*a[1:]) == 0
 
+    @mark_cuda_test
+    def test_fp_textures(self):
+        if drv.Context.get_device().compute_capability() < (1, 3):
+            return
+
+        for tp in [numpy.float32, numpy.float64]:
+            from pycuda.tools import dtype_to_ctype
+
+            tp_cstr = dtype_to_ctype(tp)
+            mod = SourceModule("""
+            #include <pycuda-helpers.hpp>
+
+            texture<fp_tex_%(tp)s, 1, cudaReadModeElementType> my_tex;
+
+            __global__ void copy_texture(%(tp)s *dest)
+            {
+              int i = threadIdx.x;
+              dest[i] = fp_tex1Dfetch(my_tex, i);
+            }
+            """ % {"tp": tp_cstr})
+
+            copy_texture = mod.get_function("copy_texture")
+            my_tex = mod.get_texref("my_tex")
+
+            import pycuda.gpuarray as gpuarray
+
+            shape = (384,)
+            a = numpy.random.randn(*shape).astype(tp)
+            a_gpu = gpuarray.to_gpu(a)
+            a_gpu.bind_to_texref_ext(my_tex, allow_double_hack=True)
+
+            dest = numpy.zeros(shape, dtype=tp)
+            copy_texture(drv.Out(dest),
+                    block=shape+(1,1,),
+                    texrefs=[my_tex])
+
+            assert la.norm(dest-a) == 0
+
+    @mark_cuda_test
+    def test_constant_memory(self):
+        # contributed by Andrew Wagner
+
+        module = SourceModule("""
+        __constant__ float const_array[32];
+
+        __global__ void copy_constant_into_global(float* global_result_array)
+        {
+            global_result_array[threadIdx.x] = const_array[threadIdx.x];
+        }
+        """)
+
+        copy_constant_into_global = module.get_function("copy_constant_into_global")
+        const_array, _ = module.get_global('const_array')
+
+        host_array = numpy.random.randint(0,255,(32,)).astype(numpy.float32)
+
+        global_result_array = drv.mem_alloc_like(host_array)
+        drv.memcpy_htod(const_array, host_array)
+
+        copy_constant_into_global(
+                global_result_array,  
+                grid=(1, 1), block=(32, 1, 1))
+
+        host_result_array = numpy.zeros_like(host_array)
+        drv.memcpy_dtoh(host_result_array, global_result_array)
+
+        assert (host_result_array == host_array).all
+
 
 
 if __name__ == "__main__":
-    unittest.main()
+    # make sure that import failures get reported, instead of skipping the tests.
+    import pycuda.autoinit
+
+    import sys
+    if len(sys.argv) > 1:
+        exec sys.argv[1]
+    else:
+        from py.test.cmdline import main
+        main([__file__])
